from ._burst_interactions import global_burst_link
from ._burst_interactions import global_uni_burst_link
from ._burst_interactions import genepair_inference
from ._burst_interactions import genepair_burstinference
from ._burst_interactions import genepair_interactionsinference
from ._burst_interactions import ks_2samp
from ._downstream_analysis import tf_tg_analysis
from ._downstream_analysis import interaction_burst_regression
from ._downstream_analysis import burst_interaction_overall
from ._downstream_analysis import burst_info_summarize
from ._downstream_analysis import affinity_burst
from ._downstream_analysis import burst_interactionlevel_positive
from ._downstream_analysis import comparison_burst_analysis
from ._downstream_analysis import comparison_regulation_analysis
from ._downstream_analysis import comparison_regulation_difference_analysis
from ._downstream_analysis import go_enrichment_analysis
from ._downstream_analysis import differential_tg_GO


# -*- coding: utf-8 -*-
import os, math
import numpy as np
import pandas as pd
from scipy import stats
from scipy.stats import pearsonr
from scipy.ndimage import gaussian_filter1d, gaussian_filter
import statsmodels.api as sm
from sklearn.metrics import roc_curve, roc_auc_score
from sklearn.preprocessing import MinMaxScaler
import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
from matplotlib.colors import TwoSlopeNorm
import seaborn as sns

from _utils import _synthetic_data
from tools import _burst_interactions as bi


def landscape(params, fig_index):
    # params = [kon1, kon2, koff1, koff2, ksyn1, ksyn2, kdeg, rK, h1, h2, eps_, T]
    S_stable, mkonx, mkony, mbf1, mbf2, mbs1, mbs2, mbf1_, mbf2_, mbs1_, mbs2_ = _synthetic_data.SSA_coexpression(params, verbose = 'burst', fig = False)
    simul_data = S_stable.astype(int)
    pxy = bi.density_estimation(simul_data)
    geneinfo = np.array(['X1', 'X2', -math.copysign(1, params[8]), -math.copysign(1, params[9])]).reshape([1, 4])
    infer_result = bi.genepair_inference(simul_data, geneinfo, 0, verbose1 = False, verbose2 = False, verbose3 = 0, test = False)
    stat_params = infer_result[0: 7]
    est_pxy, est_px, est_py = bi.Binary_PoBe(stat_params, simul_data)
    fig, ax = plt.subplots(dpi=900)
    g = sns.jointplot(x=simul_data[0, :], y=simul_data[1, :], kind="kde", marginal_kws={'fill': False, 'alpha': 0.0}, fill=True, cut=0)
    px = ax.hist(simul_data[0, :], bins = np.max(simul_data[0, :]), density=True, alpha=0)[0]
    py = ax.hist(simul_data[1, :], bins = np.max(simul_data[1, :]), density=True, alpha=0)[0]
    px = gaussian_filter1d(px, sigma=0.5)
    py = gaussian_filter1d(py, sigma=0.5)
    x_range = np.arange(len(px))
    y_range = np.arange(len(py))
    g.ax_marg_x.plot(x_range, px)
    g.ax_marg_y.plot(py, y_range)
    g.ax_marg_x.fill_between(x_range, px, color='#257AB6', alpha=0.15)  
    g.ax_marg_y.fill_betweenx(y_range, py, color='#257AB6', alpha=0.15)
    g.ax_marg_x.scatter(np.arange(np.max(simul_data[0, :])), est_px, facecolors='none', edgecolors='#257AB6', s=80, linewidths=2, alpha=0.65)
    g.ax_marg_y.scatter(est_py, np.arange(np.max(simul_data[1, :])), facecolors='none', edgecolors='#257AB6', s=80, linewidths=2, alpha=0.65)
    g.ax_joint.set_xlim([0, np.max(simul_data[0, :])])
    g.ax_joint.set_ylim([0, np.max(simul_data[1, :])])
    g.ax_marg_x.set_xlim([0, np.max(simul_data[0, :])])
    g.ax_marg_y.set_ylim([0, np.max(simul_data[1, :])])
    plt.show()
    fig = plt.figure(dpi=900)
    ax = fig.add_subplot(projection='3d')
    ax.grid(False)
    ax.xaxis.pane.fill = False
    ax.yaxis.pane.fill = False
    ax.zaxis.pane.fill = False
    xx = np.arange(np.max(simul_data[0, :]))
    yy = np.arange(np.max(simul_data[1, :]))
    x, y = np.meshgrid(xx, yy)
    z = pxy
    ax.plot_surface(x, y, z, rstride=1, cstride=1, cmap=plt.cm.coolwarm)
    plt.xticks([])
    plt.yticks([])
    ax.set_zticks([])
    plt.show()
    fig = plt.figure(dpi=900)
    ax = fig.add_subplot(projection='3d')
    ax.grid(False)
    ax.xaxis.pane.fill = False
    ax.yaxis.pane.fill = False
    ax.zaxis.pane.fill = False
    xx = np.arange(np.max(simul_data[0, :]))
    yy = np.arange(np.max(simul_data[1, :]))
    x, y = np.meshgrid(xx, yy)
    z = gaussian_filter(pxy, sigma=2)
    ax.plot_surface(x, y, z, rstride=1, cstride=1, cmap=plt.cm.plasma, alpha=0.95)
    plt.xticks([])
    plt.yticks([])
    ax.set_zticks([])  
    plt.show()
    return

def ROC_plot(readfile_name):
    filtered_inference_result = filter_inference_results(readfile_name, [-0.5, 1.5, 0.0, 1.0])
    interaction_grn = np.asarray(np.vstack([filtered_inference_result[:, 2].astype(float), filtered_inference_result[:, 3].astype(float)]))
    interaction_infer = np.asarray(np.vstack([np.abs(filtered_inference_result[:, 22].astype(float)), np.abs(filtered_inference_result[:, 23].astype(float))]))
    fpr, tpr, thresholds = roc_curve(interaction_grn, interaction_infer)
    auc_score = roc_auc_score(interaction_grn, interaction_infer)
    # Create ROC plot
    fig, ax = plt.subplots(dpi=900)
    plt.plot(fpr, tpr, label='ROC Curve (AUC = {:.2f})'.format(auc_score), c = '#4292c5')
    plt.plot([0, 1], [0, 1], 'k--', label='Random', c = '#9dc4e6')
    plt.xlabel('False Positive Rate')
    plt.ylabel('True Positive Rate')
    plt.title('Receiver Operating Characteristic (ROC) Curve')
    plt.legend()
    plt.show()
    return

def filter_inference_results(readfile_name, thresholds):
    inference_result_ = pd.read_csv(readfile_name)
    inference_result = np.matrix(inference_result_)[:, 1::]
    # filter out
    indices_infer = np.where(inference_result[:, 17].astype(float) == 0)[0]
    for index in np.arange(4):
        indices_ = np.where(inference_result[:, index+4].astype(float) == 1e-4)[0]
        indices__ = np.where(inference_result[:, index+4].astype(float) == 1e4)[0]
        indices___ = list(set(indices_) | set(indices__))
        indices_infer = list(set(indices_infer) | set(indices___))
    for index in np.arange(2):
        indices_ = np.where(np.log10(inference_result[:, index+11].astype(float)) < thresholds[0])[0]
        indices__ = np.where(np.log10(inference_result[:, index+11].astype(float)) > thresholds[1])[0]
        indices___ = list(set(indices_) | set(indices__))
        indices_infer = list(set(indices_infer) | set(indices___))
    for index in np.arange(2):
        indices_ = np.where(np.log10(inference_result[:, index+13].astype(float)) < thresholds[2])[0]
        indices__ = np.where(np.log10(inference_result[:, index+13].astype(float)) > thresholds[3])[0]
        indices___ = list(set(indices_) | set(indices__))
        indices_infer = list(set(indices_infer) | set(indices___))
    filtered_inference_result = np.delete(inference_result, indices_infer, axis = 0)
    return(filtered_inference_result)

def network_umap_pre(read_filename, save_filename):
    filtered_inference_result = filter_inference_results(read_filename, [-2.5, 2.0, -1.5, 2.0])
    result_ = np.vstack([filtered_inference_result[:, np.array([0, 1, 11, 13, 15, 21, 23])], filtered_inference_result[:, np.array([1, 0, 12, 14, 16, 20, 22])]])
    sorted_result_ = np.matrix(sorted(result_.tolist(), key = lambda x: (x[0], x[1])))
    # sorted_result_ = [genename_target, genename_tf, bf, bs, cv2, cmi, feedback_infer]
    index_interaction = np.where(sorted_result_[:, 6].astype(float) == 0)[0]
    sorted_result = np.delete(sorted_result_, index_interaction, axis = 0)
    unique_genename = sorted(list(set(np.vstack([np.array(sorted_result[:, 0]), np.array(sorted_result[:, 1])]).flatten().tolist())))
    counts_data_ = pd.read_csv(os.path.abspath('readfiles/scRNA-seq/countsdata matrix.csv'))
    counts_data = np.delete(np.matrix(counts_data_.values.tolist()), np.s_[:1], axis=1)
    counts_index = np.where(np.isin(counts_data[:, 0], unique_genename))[0]
    counts = counts_data[counts_index, 1::].astype(float)
    row_mean = np.nanmean(counts, axis=1)
    inds = np.where(np.isnan(counts))
    counts[inds] = np.take(row_mean, inds[0])
    counts = np.hstack([np.asarray(unique_genename).reshape([len(unique_genename), 1]), counts])
    df = pd.DataFrame(counts)
    df.to_csv(save_filename) 
    return (counts)

def network_pre(readfile_name, savefile_name):
    filtered_inference_result = filter_inference_results(readfile_name, [-3.5, 2.5, -2.0, 3.0])
    result_ = np.vstack([np.array(filtered_inference_result[:, np.array([1, 0, 23, 21])]), np.array(filtered_inference_result[:, np.array([0, 1, 22, 20])])])
    sorted_result = np.matrix(sorted(result_.tolist(), key = lambda x: (x[0], x[1])))
    unique_genename = sorted(list(set(np.array(sorted_result[:, 0]).flatten().tolist())))
    indices = np.where(sorted_result[:, 2].astype(float) == 0)[0]
    sorted_result = np.delete(sorted_result, indices, axis = 0)
    feedback = sorted_result.copy()
    index = 1
    for genename in unique_genename:
        indices_ = np.where(sorted_result[:, 0] == genename)[0]
        feedback[indices_, 0] = index
        indices__ = np.where(np.isin(unique_genename, sorted_result[indices_, 1]))[0]
        feedback[indices_, 1] = indices__ + 1
        index = index + 1
    df = pd.DataFrame(feedback)
    df.to_csv(savefile_name) 
    return(feedback)

def burst_pre(readfile_name, savefile_name):
    filtered_inference_result = filter_inference_results(readfile_name, [-2.5, 2.0, -1.5, 2.0])
    result_ = np.vstack([filtered_inference_result[:, np.array([0, 11, 13, 15, 21, 23])], filtered_inference_result[:, np.array([1, 12, 14, 16, 20, 22])]])
    sorted_result = np.matrix(sorted(result_.tolist(), key = lambda x: (x[0])))
    # sorted_result = [genename, bf, bs, cv2, cmi, feedback_infer]
    unique_genename = sorted(list(set(np.array(sorted_result[:, 0]).flatten().tolist())))
    result_unique_ = np.empty([len(unique_genename), 5])
    # result_unique_ = [bf, bs, cv2, expression_level, feedback]
    index = 0
    for genename in unique_genename:
        indices_ = np.where(sorted_result[:, 0] == genename)[0]
        if (len(indices_) == 1):
            for n in np.arange(3):
                result_unique_[index, n] = sorted_result[indices_, n+1].astype(float)
            result_unique_[index, 3] = result_unique_[index, 0] * result_unique_[index, 1]
            result_unique_[index, 4] = sorted_result[indices_, 4].astype(float) * sorted_result[indices_, 5].astype(float)
        elif (len(indices_) > 1):
            for n in np.arange(3):
                data = np.array(sorted_result[indices_, n+1].astype(float)).flatten()
                z_scores = np.abs(stats.zscore(data))
                filtered_data = np.array(data)[z_scores < 2]
                result_unique_[index, n] = np.mean(filtered_data)
            result_unique_[index, 3] = result_unique_[index, 0] * result_unique_[index, 1]
            result_unique_[index, 4] = np.sum(np.array(sorted_result[indices_, 4].astype(float)).flatten() * np.array(sorted_result[indices_, 5].astype(float)).flatten())
        index = index + 1
    df = pd.DataFrame(result_unique_)
    df.to_csv(savefile_name) 
    return(result_unique_)

def barplot(vals, ylim, savefile_name):
    fig, ax = plt.subplots(dpi=900)
    categories = ['A', 'B']
    plt.bar(categories, vals)
    plt.title('Sample Bar Plot')
    plt.xlabel('Categories')
    plt.ylabel('Values')
    plt.ylim(ylim[0], ylim[1])
    plt.savefig(savefile_name)     
    plt.show()
    return

def gene_roles_heatmap(vals):
    data = np.asarray(np.vstack([vals[:, 0].astype(float), vals[:, 1].astype(float)]))
    scaler = MinMaxScaler()
    normalized_data = scaler.fit_transform(data.reshape(-1, 1))
    scaled_transformed = np.power(normalized_data, 2) 
    degree_data = normalized_data.reshape([vals.shape[1], vals.shape[0]])
    plt.figure(dpi=900)
    # Create a heatmap
    sns.heatmap(degree_data, annot=False, annot_kws={'size': 100}, cmap=plt.cm.Greens, square=True, cbar=False, xticklabels=False, yticklabels=False,)
    plt.xlabel('Gene roles', fontsize=3) 
    plt.title('Identification of gene roles in gene regulatory intercations network', fontsize=5, fontweight='bold')  
    plt.show()
    return

def subnetwork_pre(readfile_name, genename_, savefile_name):
    filtered_inference_result = filter_inference_results(readfile_name, [-3.5, 2.5, -2.0, 3.0])
    result_all = np.vstack([np.array(filtered_inference_result[:, np.array([1, 0, 23, 21])]), np.array(filtered_inference_result[:, np.array([0, 1, 22, 20])])])
    unique_genename_all = sorted(list(set(np.array(result_all[:, 0]).flatten().tolist())))
    indices1 = np.where(filtered_inference_result[:, 0] == genename_)[0]
    indices2 = np.where(filtered_inference_result[:, 1] == genename_)[0]
    indices = list(set(indices1) | set(indices2))
    filtered_inference_result = filtered_inference_result[indices, :]
    result_ = np.vstack([np.array(filtered_inference_result[:, np.array([1, 0, 23, 21])]), np.array(filtered_inference_result[:, np.array([0, 1, 22, 20])])])
    sorted_result = np.matrix(sorted(result_.tolist(), key = lambda x: (x[0], x[1])))
    indices = np.where(sorted_result[:, 2].astype(float) == 0)[0]
    sorted_result = np.delete(sorted_result, indices, axis = 0)
    unique_genename = sorted(list(set(np.array(sorted_result[:, 0]).flatten().tolist())))
    genename_index = np.where(np.isin(unique_genename_all, unique_genename))[0] + 1
    # df = pd.DataFrame(genename_index)
    # df.to_csv(savefile_name) 
    return(genename_index)

def scatter_plot(x, y, x_sd, y_sd):
    x_err = x_sd
    y_err = y_sd
    colors = ['blue', 'orange', 'grey', 'blue', 'orange', 'grey']
    categories = ['negative for low-expression', 'positive for low-expression', 'none for low-expression', 'negative for high-expression', 'positive for high-expression', 'none for high-expression']
    plt.rcParams.update({
        'font.size': 3, 'axes.labelsize': 3, 'axes.titlesize': 4,       
        'xtick.labelsize': 4,'ytick.labelsize': 4,         #
        'legend.fontsize': 2,       
        'lines.linewidth': 0.4,      
        'lines.markersize': 2, 'errorbar.capsize': 0.9,       
        'grid.linewidth': 0.4,       
        'figure.dpi': 300, 'figure.figsize': (2, 1.5)})
    fig, ax = plt.subplots()
    for i in range(3):
        plt.errorbar(x[i], y[i], xerr=x_err[i], yerr=y_err[i], fmt='o',  color=colors[i], markersize=1.8, capsize=1, capthick=1, label=categories[i])
    for i in range(3, len(x)):
        plt.errorbar(x[i], y[i], xerr=x_err[i], yerr=y_err[i], fmt='s',  color=colors[i], markersize=1.8, capsize=1, capthick=1, label=categories[i])
    for spine in ax.spines.values():
        spine.set_linewidth(0.5)  
    ax.tick_params(axis='both', which='major', width=0.4) 
    ax.tick_params(axis='both', which='minor', width=0.4) 
    plt.xlabel('log10(bf)')
    plt.ylabel('log10(cv2)')
    plt.grid(True) 
    legend_handles = [
        Line2D([0], [0], marker='o', color='w', markerfacecolor='blue', markersize=1.5, label='negative for low-expression'),
        Line2D([0], [0], marker='o', color='w', markerfacecolor='orange', markersize=1.5, label='positive for low-expression'),
        Line2D([0], [0], marker='o', color='w', markerfacecolor='grey', markersize=1.5, label='none for low-expression'),
        Line2D([0], [0], marker='s', color='w', markerfacecolor='blue', markersize=1.5, label='negative for high-expression'),
        Line2D([0], [0], marker='s', color='w', markerfacecolor='orange', markersize=1.5, label='positive for high-expression'),
        Line2D([0], [0], marker='s', color='w', markerfacecolor='grey', markersize=1.5, label='none for high-expression')]
    plt.legend(handles=legend_handles, loc='best', handletextpad=0.5, fontsize=2)
    plt.show()
    return

def box_plot3(data_bf1, data_bf2, data_bs1, data_bs2, data_cv21, data_cv22):
    data_bf = [data_bf1, data_bf2]
    data_bs = [data_bs1, data_bs2]
    data_cv2 = [data_cv21, data_cv22]
    data = [data_bf, data_bs, data_cv2]
    fig, ax = plt.subplots(dpi=900)
    positions = [[1, 2], [4, 5], [7, 8]] 
    colors = ['skyblue', 'peachpuff']  
    for i, group_data in enumerate(data):
        for j, group in enumerate(group_data):
            ax.boxplot(group, positions=[positions[i][j]], patch_artist=True,
                       boxprops=dict(facecolor=colors[j], linewidth=3),
                       capprops=dict(linewidth=1.0), whiskerprops=dict(linewidth=1.0),  
                       medianprops=dict(linewidth=0.9, color='red'),  
                       showfliers=False)
    ax.set_xticks([1.5, 4.5, 7.5])
    ax.set_xticklabels(['BF', 'BS', 'CV2'], fontsize=3)
    ax.set_ylabel('log10(Value)', fontsize=3)
    plt.show()
    t_stat_bf, p_value_bf = stats.ttest_ind(data_bf1, data_bf2)
    t_stat_bs, p_value_bs = stats.ttest_ind(data_bs1, data_bs2)
    t_stat_cv2, p_value_cv2 = stats.ttest_ind(data_cv21, data_cv22)
    return(p_value_bf, p_value_bs, p_value_cv2)

def box_plot3_2(data1_bf1, data1_bf2, data1_bs1, data1_bs2, data1_cv21, data1_cv22, data2_bf1, data2_bf2, data2_bs1, data2_bs2, data2_cv21, data2_cv22):
    data1_bf = [data1_bf1, data1_bf2]
    data1_bs = [data1_bs1, data1_bs2]
    data1_cv2 = [data1_cv21, data1_cv22]
    data1 = [data1_bf, data1_bs, data1_cv2]
    data2_bf = [data2_bf1, data2_bf2]
    data2_bs = [data2_bs1, data2_bs2]
    data2_cv2 = [data2_cv21, data2_cv22]
    data2 = [data2_bf, data2_bs, data2_cv2]
    fig, axes = plt.subplots(1, 2, figsize=(4, 1.5), dpi=300) 
    positions = [[1, 2], [4, 5], [7, 8]] 
    colors = ['skyblue', 'peachpuff']  
    for i, group_data in enumerate(data1):
        for j, group in enumerate(group_data):
            axes[0].boxplot(group, positions=[positions[i][j]], patch_artist=True,
                       widths=0.6,
                       boxprops=dict(facecolor=colors[j], linewidth=0.5),
                       capprops=dict(linewidth=0.5), whiskerprops=dict(linewidth=0.3),  
                       medianprops=dict(linewidth=0.5, color='red'),  
                       showfliers=False)
    axes[0].set_xticks([1.5, 4.5, 7.5])
    axes[0].set_xticklabels(['BF', 'BS', 'CV2'], fontsize=6)
    axes[0].set_ylabel('log10(Value)', fontsize=5.5)
    axes[0].tick_params(axis='y', labelsize=5)
    for spine in axes[0].spines.values():
        spine.set_linewidth(0.5) 
    axes[0].tick_params(axis='both', width=0.5) 
    for i, group_data in enumerate(data2):
        for j, group in enumerate(group_data):
            axes[1].boxplot(group, positions=[positions[i][j]], patch_artist=True,
                       widths=0.6,
                       boxprops=dict(facecolor=colors[j], linewidth=0.5),
                       capprops=dict(linewidth=0.5), whiskerprops=dict(linewidth=0.3),  
                       medianprops=dict(linewidth=0.5, color='red'),  
                       showfliers=False)
    axes[1].set_xticks([1.5, 4.5, 7.5])
    axes[1].set_xticklabels(['BF', 'BS', 'CV2'], fontsize=6)
    axes[1].set_ylabel('log10(Value)', fontsize=5.5)
    axes[1].tick_params(axis='y', labelsize=5)
    for spine in axes[1].spines.values():
        spine.set_linewidth(0.5) 
    axes[1].tick_params(axis='both', width=0.5) 
    plt.tight_layout()
    plt.show()
    t_stat1_bf, p_value1_bf = stats.ttest_ind(data1_bf1, data1_bf2)
    t_stat1_bs, p_value1_bs = stats.ttest_ind(data1_bs1, data1_bs2)
    t_stat1_cv2, p_value1_cv2 = stats.ttest_ind(data1_cv21, data1_cv22)
    t_stat2_bf, p_value2_bf = stats.ttest_ind(data2_bf1, data2_bf2)
    t_stat2_bs, p_value2_bs = stats.ttest_ind(data2_bs1, data2_bs2)
    t_stat2_cv2, p_value2_cv2 = stats.ttest_ind(data2_cv21, data2_cv22)
    return(p_value1_bf, p_value1_bs, p_value1_cv2, p_value2_bf, p_value2_bs, p_value2_cv2)

def fit_3d(x, y, z):
    X = np.column_stack((x, y))
    X = sm.add_constant(X)  
    model = sm.OLS(z, X).fit()  
    intercept, a, b = model.params
    p_values = model.pvalues
    return(intercept, a, b, p_values[0], p_values[1], p_values[2])

def scatter_2d(x, y, color, x_label, y_label):
    fig, ax = plt.subplots(dpi=900)
    sc = plt.scatter(x, y, s=35, c=color, cmap=plt.cm.viridis)
    plt.colorbar(sc)
    plt.xlabel(x_label, fontsize=17)
    plt.ylabel(y_label, fontsize=17)         
    # plt.show()
    plt.clf()
    t_stat, p_value = stats.ttest_ind(x, y)
    return(p_value)

def scatter_interval(data):
    x1 = np.log10(data[1::, 0].astype(float))
    y1 = np.log10(data[1::, 2].astype(float))
    x2 = np.log10(data[1::, 1].astype(float))
    y2 = np.log10(data[1::, 3].astype(float))
    corr_coef1, p_value1 = pearsonr(x1, y1)
    corr_coef2, p_value2 = pearsonr(x2, y2)

    fig, axes = plt.subplots(1, 2, figsize=(4, 1.5), dpi=300) 
    sns.regplot(x=x1, y=y1, ci=95, scatter_kws={'alpha': 0.8, 's': 12}, line_kws={'linewidth': 0.8}, ax=axes[0])
    axes[0].set_xlabel('log10(bf_Fibr.)', fontsize=3.5)
    axes[0].set_ylabel('log10(bf_ES)', fontsize=3.5)
    axes[0].set_xlim(-1.2, 1.7)
    axes[0].set_ylim(-1.2, 1.7)
    axes[0].tick_params(axis='both', which='both', labelsize=3, width=0.3)
    sns.regplot(x=x2, y=y2, ci=95, scatter_kws={'alpha': 0.8, 's': 12}, line_kws={'linewidth': 0.8}, ax=axes[1])
    axes[1].set_xlabel('log10(bs_Fibr.)', fontsize=3.5)
    axes[1].set_ylabel('log10(bs_ES)', fontsize=3.5)
    axes[1].set_xlim(-0.7, 1.5)
    axes[1].set_ylim(-0.7, 1.5)
    axes[1].tick_params(axis='both', which='both', labelsize=3, width=0.3)
    for ax in axes:
        for spine in ax.spines.values(): spine.set_linewidth(0.4)
    plt.tight_layout()   
    plt.show()
    return(p_value1, p_value2)

def box_plot4_2(bf, bs, cv2, mean):
    fig, ax = plt.subplots(1, 1, figsize=(3, 1), dpi=300)
    positions = [[1, 2], [4, 5], [7, 8], [10, 11]]  
    colors = ['#90EE90', '#FFDAB9']
    for i, group_data in enumerate([bf, bs, cv2, mean]): 
        for j, subgroup in enumerate(group_data): 
            ax.boxplot(subgroup, positions=[positions[i][j]], patch_artist=True, widths=0.6, boxprops=dict(facecolor=colors[j], linewidth=0.5),
                     capprops=dict(linewidth=0.5), whiskerprops=dict(linewidth=0.3), medianprops=dict(linewidth=0.5, color='red'), showfliers=False)
    ax.set_xticks([1.5, 4.5, 7.5, 10.5])
    ax.set_xticklabels(['BF', 'BS', 'CV2', 'Mean'], fontsize=4.5)
    ax.set_ylim([-1.4, 2.8])
    ax.set_ylabel('log10(Value)', fontsize=4)
    ax.tick_params(axis='y', labelsize=4)
    for spine in ax.spines.values():
        spine.set_linewidth(0.5)
    ax.tick_params(axis='both', width=0.5)
    plt.show()
    return

def heatmap_regulation(data):
    plt.figure(figsize=(5, 0.9), dpi=300) 
    vmin = np.min(data)
    vmax = np.max(data)
    vcenter = 0
    if vmin >= vcenter: vmin = vcenter - 0.1  
    if vmax <= vcenter: vmax = vcenter + 0.1  
    sns.heatmap(data, cmap=plt.cm.coolwarm , norm=TwoSlopeNorm(vmin=vmin, vcenter=vcenter, vmax=vmax), xticklabels=False, yticklabels=False, linewidths=0.001, linecolor='white', cbar=False)  
    plt.title('Comparison of the overall regulatory interaction levles between two groups', fontsize=6)
    plt.tight_layout()
    plt.show()
    return

def barplot_regulation(data):
    plt.figure(dpi=900, figsize=(4, 0.95))
    colors = ['#FF7F7F' if i < 16 else 'skyblue' if i >= 88 else 'gray' for i in range(144)]
    plt.bar(np.arange(1, data.shape[0]+1), data, width=0.9, color=colors, alpha=0.8)
    plt.title('Difference values of the overall regulatory interaction levles between two groups', fontsize=5)
    plt.xlabel('Genename', fontsize=5)
    plt.ylabel('Values', fontsize=5)
    plt.grid(axis='y', alpha=0.4)
    plt.xticks([])
    plt.yticks([])
    ax = plt.gca()
    for spine in ax.spines.values(): spine.set_linewidth(0.35) 
    plt.show()
    return

def visualize_go_bar(go_results, title, color_):
    go_results = go_results.sort_values("Adjusted P-value").head(10)
    palette = sns.color_palette(color_, n_colors=10)
    fig, ax = plt.subplots(figsize=(4.5, 3.5), dpi=300)
    barplot = sns.barplot(
        x=go_results["Overlap"].apply(lambda x: int(x.split('/')[0])),  
        y=go_results["Term"], palette=palette, hue=None, ax=ax, legend=False)
    norm = mpl.colors.Normalize(vmin=go_results["Adjusted P-value"].min(), vmax=go_results["Adjusted P-value"].max())
    sm = plt.cm.ScalarMappable(cmap=color_, norm=norm) 
    sm.set_array([])
    cbar = fig.colorbar(sm, ax=ax, orientation="vertical", pad=0.02)
    cbar.set_label("p-value", fontsize=4)
    ax.set_title(title, fontsize=10, weight='bold', pad=15)
    ax.set_xlabel("Gene number", fontsize=8)
    ax.set_ylabel("GO terms", fontsize=8)
    plt.tight_layout()
    plt.show()
    return


# -*- coding: utf-8 -*-
import numpy as np
import pandas as pd
import pyarrow.parquet as pq
import matplotlib.pyplot as plt


def select_genepair_grn(read_filename, save_filename, verbose):
    """ Select gene-pair interactions from base-GRN. """
    parquet_file = pq.ParquetFile(read_filename)
    base_GRN = parquet_file.read().to_pandas()
    tf_genename = base_GRN.columns.tolist()[2::]
    target_genename = base_GRN['gene_short_name'].tolist()
    tf_genename_lower = [s.lower() for s in tf_genename]
    target_genename_lower = [s.lower() for s in target_genename]
    base_grn = np.delete(np.matrix(base_GRN), np.s_[:2], axis=1)
    grn = find_genepair(tf_genename_lower, target_genename_lower, base_grn)
    if (verbose == True):
        df = pd.DataFrame(grn)
        df.to_csv(save_filename) 
        return (grn)
    else: return (grn)

def find_genepair(arr1, arr2, grn_matrix):
    grn = np.array(['gene1', 'gene2', 'feedback1', 'feedback2']).reshape([1, 4])
    genename = list(set(arr1) & set(arr2))
    indices1 = np.where(np.isin(arr1, genename))[0]
    indices2 = np.where(np.isin(arr2, genename))[0]
    for index1 in indices1:
        str1 = arr1[index1]
        for index2 in indices2:
            str2 = arr2[index2]
            index11 = arr2.index(str1)    
            index22 = arr1.index(str2)  
            grn_ = np.array([str1, str2, grn_matrix[index2, index1], grn_matrix[index11, index22]]).reshape([1, 4])
            grn = np.vstack([grn, grn_])        
    return (grn)

def integration_grn(read_filename_atac, read_filename_tf, save_filename, verbose):
    """ Integration for grn from different sources. """
    grn_atac = pd.read_csv(read_filename_atac)
    grn_tf = pd.read_csv(read_filename_tf)
    grn_atac = np.array(grn_atac)[:, 1::]
    grn_tf = np.array(grn_tf)
    filtered_grn_atac = filtered_genepair_grn(grn_atac)
    filtered_grn_tf = filtered_genepair_grn(grn_tf)
    integrated_grn = refiltered_genepair_grn(filtered_grn_atac, filtered_grn_tf)
    if (verbose == True):
        df = pd.DataFrame(integrated_grn)
        df.to_csv(save_filename) 
        return (integrated_grn)
    else:
        return (integrated_grn)
   
def filtered_genepair_grn(grn):
    grn_ = grn.tolist()
    sorted_grn = sorted(grn_, key = lambda x: (x[0], x[1], x[2], x[3]))
    sorted_grn_ = np.matrix(sorted_grn)
    genename1 = sorted_grn_[:, 0]
    genename2 = sorted_grn_[:, 1]
    feedback1 = sorted_grn_[:, 2].astype(float)
    feedback2 = sorted_grn_[:, 3].astype(float)
    # filter for same genens & multi feedback 
    genename1_diff = genename1[0: -1] == genename1[1: :]
    indices1 = np.where(genename1_diff == True)[0]
    genename2_diff = genename2[0: -1] == genename2[1: :]
    indices2 = np.where(genename2_diff == True)[0]
    feedback1_diff = feedback1[0: -1] - feedback1[1: :]
    indices1_ = np.where(feedback1_diff == 0)[0]
    feedback2_diff = feedback2[0: -1] - feedback2[1: :]
    indices2_ = np.where(feedback2_diff == 0)[0]
    indices = list(set(indices1) & set(indices2) & set(indices1_) & set(indices2_))
    filtered_grn = np.delete(sorted_grn_, indices, axis=0)
    # filter for same genes & feedback = 1
    genename1 = filtered_grn[:, 0]
    genename2 = filtered_grn[:, 1]
    genename1_diff = genename1[0: -1] == genename1[1: :]
    indices1 = np.where(genename1_diff == True)[0]
    genename2_diff = genename2[0: -1] == genename2[1: :]
    indices2 = np.where(genename2_diff == True)[0]
    indices = list(set(indices1) & set(indices2))
    filtered_grn = np.delete(filtered_grn, indices, axis=0)
    # filter for A-A same genenames
    genename1 = filtered_grn[:, 0]
    genename2 = filtered_grn[:, 1]
    same_genename = genename1 == genename2
    indices = np.where(same_genename == True)[0]
    filtered_grn = np.delete(filtered_grn, indices, axis=0)
    # filter for [genename1, genename2] & [genename2, genename1]
    genename1 = np.array(filtered_grn[:, 0]).flatten().tolist()
    unique_genename1 = list(set(genename1))
    sorted_unique_genename1 = sorted(unique_genename1)
    for uni_genename1 in sorted_unique_genename1:
        genename1 = filtered_grn[:, 0]
        genename2 = filtered_grn[:, 1]
        indices1 = np.where(np.isin(genename1, uni_genename1))[0]
        indices2 = np.where(np.isin(genename2, uni_genename1))[0]
        indices_ = np.where(np.isin(genename1[indices2], genename2[indices1]))[0]
        indices = indices2[indices_]
        filtered_grn = np.delete(filtered_grn, indices, axis=0)  
    return (filtered_grn)
  
def refiltered_genepair_grn(filtered_grn_atac, filtered_grn_tf):
    genename1_atac = filtered_grn_atac[:, 0]
    genename2_atac = filtered_grn_atac[:, 1]
    genename1_tf = filtered_grn_tf[:, 0]
    genename2_tf = filtered_grn_tf[:, 1]
    intersection1 = list(set(np.array(genename1_atac).flatten().tolist()) & set(np.array(genename1_tf).flatten().tolist()))
    intersection2 = list(set(np.array(genename2_atac).flatten().tolist()) & set(np.array(genename2_tf).flatten().tolist()))
    indices1 = np.where(np.isin(genename1_atac, intersection1))[0]
    indices2 = np.where(np.isin(genename2_atac, intersection2))[0]
    indices = list(set(indices1) & set(indices2))
    refiltered_grn_atac = np.delete(filtered_grn_atac, indices, axis=0)
    integrated_grn = np.vstack([refiltered_grn_atac, filtered_grn_tf]) 
    sorted_integrated_grn = np.matrix(sorted(integrated_grn.tolist(), key = lambda x: (x[0], x[1], x[2], x[3])))
    return (sorted_integrated_grn)
   
def RNAseq_analysis(read_filename1, read_filename2, save_filename1, save_filename2, verbose):
    """ scRNA-seq raw data preprocess. """
    df1 = pd.read_csv(read_filename1)
    df2 = pd.read_csv(read_filename2)
    rnaseq_list1 = df1.values.tolist()
    rnaseq_list2 = df2.values.tolist()
    rnaseq_matrix1 = np.delete(np.matrix(rnaseq_list1), np.s_[:1], axis=1)
    rnaseq_matrix2 = np.delete(np.matrix(rnaseq_list2), np.s_[:1], axis=1)
    genename1 = [row[0] for row in rnaseq_list1]
    genename1 = [s.lower() for s in genename1]
    genename2 = [row[0] for row in rnaseq_list2]
    genename2 = [s.lower() for s in genename2]
    if (genename1 == genename2):
        genename = genename1
        mm = len(genename)
        counts = np.hstack((rnaseq_matrix1, rnaseq_matrix2)).astype(float)
    else:
        print('genenames of two matrices can not match')

    stat = np.array(['mean', 'var', 'cv2']).reshape([1, 3])
    for index in np.arange(mm):
        countdata = counts[index, :]
        countdata_without_nan = countdata[~np.isnan(countdata)]
        mean_ = np.mean(countdata_without_nan)
        var_ = np.var(countdata_without_nan)
        cv2_ = var_ / (mean_**2)
        stat_ = np.array([mean_, var_, cv2_]).reshape([1, 3])
        stat = np.vstack([stat, stat_])
    
    mean = stat[1::, 0].astype(float)
    fig1, ax1 = plt.subplots(dpi=900)
    plt.hist(mean, bins=20, density=True)
    # plt.savefig('rawdata_mean.pdf')  
    plt.show()
    fig2, ax2 = plt.subplots(dpi=900)
    plt.hist(np.log(mean), bins=20, density=True)
    # plt.savefig('rawdata_logmean.pdf')  
    plt.show()
    fig3, ax3 = plt.subplots(dpi=900)
    df_mean = pd.DataFrame(np.log(mean))
    plt.boxplot(df_mean[0])
    # plt.savefig('rawdata_boxplot.pdf')  
    plt.show()
    
    counts_matrix = np.hstack([np.matrix(genename).reshape([mm, 1]), counts])
    stat_matrix = np.hstack([np.matrix(genename).reshape([mm, 1]), np.matrix(stat[1::, :]).astype(float)])
    if (verbose == True):
        df1 = pd.DataFrame(counts_matrix)
        df1.to_csv(save_filename1) 
        df2 = pd.DataFrame(stat_matrix)
        df2.to_csv(save_filename2) 
        return(counts_matrix, stat_matrix)
    else: return(counts_matrix, stat_matrix)

def RNAseq_analysis_uni(read_filename, save_filename, verbose):
    genename = np.matrix(pd.read_csv(read_filename))[:, 1]
    counts = np.matrix(pd.read_csv(read_filename))[:, 2::].astype(float)
    stat = np.array(['mean', 'var', 'cv2']).reshape([1, 3])
    for index in np.arange(counts.shape[0]):
        countdata = counts[index, :]
        mean_ = np.mean(countdata)
        if (mean_ == 0): mean_ = 1e-6
        var_ = np.var(countdata)
        cv2_ = var_ / (mean_**2)
        stat = np.vstack([stat, np.array([mean_, var_, cv2_]).reshape([1, 3])])
    mean = stat[1::, 0].astype(float)
    filtered_mean = np.delete(mean, np.where(mean == 1e-6)[0])

    fig1, ax1 = plt.subplots(dpi=900)
    plt.hist(np.log(filtered_mean), bins=35, density=True) 
    plt.show()
    fig2, ax2 = plt.subplots(dpi=900)
    df_mean = pd.DataFrame(np.log(filtered_mean))
    plt.boxplot(df_mean[0]) 
    plt.show()
    
    stat_matrix = np.hstack([np.matrix(genename).reshape([counts.shape[0], 1]), np.matrix(stat[1::, :]).astype(float)])
    if (verbose == True):
        df = pd.DataFrame(stat_matrix)
        df.to_csv(save_filename) 
        return(stat_matrix)
    else: return(stat_matrix)

   
def selection_GRNandRNAseq(grn_filename, rnaseq_filename, counts_filename, threshold_value, verbose):
    """ Integration with scRNA-seq data and scATAC-seq data. """
    grn = pd.read_csv(grn_filename)
    grn = np.matrix(grn)[:, 1::]
    # cluster for feedback_GRN: (1, 1)&(1, 0)&(0, 0)
    dv = grn[:, 2] - grn[:, 3]
    indices_uni_feedback = [i for i, x in enumerate(dv) if abs(x) == 1]
    grn_uni_feedback = grn[indices_uni_feedback, :]
    indices1 = [i for i, x in enumerate(dv) if x == 0]
    indices2 = [i for i, x in enumerate(grn[:, 2]) if x == 0]
    indices_no_feedback = list(set(indices1) & set(indices2))
    grn_no_feedback = grn[indices_no_feedback, :]
    indices_ = list(set(indices_no_feedback) | set(indices_uni_feedback))
    grn_bio_feedback = np.delete(grn, indices_, axis=0)

    # selection single-cell data for mean
    stat_matrix = np.matrix(pd.read_csv(rnaseq_filename))
    counts_matrix = np.matrix(pd.read_csv(counts_filename))
    indices_mean = [i for i, x in enumerate(stat_matrix[:, 2]) if x >= threshold_value]
    selected_genename = stat_matrix[indices_mean, 1]
    selected_counts_matrix = counts_matrix[indices_mean, :]
    
    # comparison between grn and single-cell
    selected_genepair_bio_feedback = comparison(grn_bio_feedback, selected_genename)
    selected_genepair_uni_feedback = comparison(grn_uni_feedback, selected_genename)
    selected_genepair_no_feedback = comparison(grn_no_feedback, selected_genename)
    selected_genepair = np.vstack([selected_genepair_bio_feedback, selected_genepair_uni_feedback, selected_genepair_no_feedback]) 

    if (verbose == True):
        df1 = pd.DataFrame(selected_genepair)
        df1.to_csv('selected_genepair.csv') 
        df2 = pd.DataFrame(selected_counts_matrix)
        df2.to_csv('selected_counts_matrix.csv') 
        return(selected_genepair, selected_counts_matrix)
    else: return(selected_genepair, selected_counts_matrix)
    
def comparison(selected_grn, rna_seq_genename):
    """ Screen for gene pair names in both scRNA-seq data and scATAC-seq data. """
    genename1 = selected_grn[:, 0]
    genename2 = selected_grn[:, 1]
    indices1 = np.where(np.isin(genename1, rna_seq_genename))[0]
    indices2 = np.where(np.isin(genename2, rna_seq_genename))[0]
    indices = list(set(indices1) & set(indices2))
    selected_genepair = selected_grn[indices, :]
    return (selected_genepair)


# -*- coding: utf-8 -*-
import os
import numpy as np
import pandas as pd
import random, math, warnings
from scipy import stats
from scipy.stats import poisson, norm, gaussian_kde, spearmanr
from scipy.special import j_roots
from scipy.special import beta as beta_fun
from scipy.optimize import minimize
from sklearn.metrics.cluster import normalized_mutual_info_score
import matplotlib.pyplot as plt
from matplotlib.transforms import Affine2D

from _utils import _synthetic_data


def global_burst_link(grn_filename, counts_filename, save_filename, verbose1, verbose2, verbose3, test, verbose4):
    """
    Inferface for inferring global dynamic information including gene regulatory interaction and bursting kinetics from scRNA-seq data of two genes.

    Parameters
    ----------
    grn_filename
        The optional input for base-GRN based on scATAC-seq.
    counts_filename
        The input for scRNA-seq data file.
    save_filename
        The saved inferred file name.
    verbose1
        Whether to output images.
        If False, only the inferred results are outputed. If True, the inferred results with visualization images are outputed.
    verbose2
        Whether to save images.
        If False, the images are only shown in the runtime screen. If True, the images are saved as pdf format in a folder under the default path.
    verbose3
        Whether to exhibit all images.
        If 0, all images are exhibited. If 1, only visualization images are exhibited.
    test
        Whether to perform distributional test.
        If False, distribution test is not performed. If True, gibbs samples are generated and distribution test is performed.
    verbose4
        Whether to save the inferred results.
        If False, the inferred results are not saved. If True, the inferred results are saved as csv format in a folder under the default path.
    
    Returns
    -------
    np.array(['gene1', 'gene2', 'feedback1', 'feedback2', 'kon1', 'kon2', 'koff1', 'koff2', 'ksyn1', 'ksyn2', 'w', 'bs1', 'bs2', 'bf1', 'bf2', 'cv1', 'cv2', 'convergence', 'corr', 'NMI', 'cmi_given_x1', 'cmi_given_x2', 'sign_givenx1', 'sign_givenx2'])
   """
    
    grn = pd.read_csv(grn_filename)
    counts = pd.read_csv(counts_filename)
    grn = np.matrix(grn)[:, 1::]
    counts = np.matrix(counts)[:, 2::]
    results =  np.array(['gene1', 'gene2', 'feedback1', 'feedback2', 'kon1', 'kon2', 'koff1', 'koff2', 'ksyn1', 'ksyn2', 'w', 'bs1', 'bs2', 'bf1', 'bf2', 'cv1', 'cv2', 'convergence', 'corr', 'NMI', 'cmi_given_x1', 'cmi_given_x2', 'sign_givenx1', 'sign_givenx2']).reshape([1, 24])
    num = len(grn[:, 0])
    for index in np.arange(num):
        try:
            geneinfo, counts_data = genepair_info(grn, counts, index)
            result_ = genepair_inference(counts_data, geneinfo, index, verbose1, verbose2, verbose3, test)
            result__ = np.hstack([geneinfo, result_.reshape([1, 20])])
            results = np.vstack([results, result__])
        except: continue
    if (verbose4 == True):
        df = pd.DataFrame(results)
        df.to_csv(save_filename) 
        return (results)
    else: return(results)
    
def global_uni_burst_link(grn_filename, counts_filename, genename1, genename2, verbose1, verbose2, verbose3, test):
    """
    Inferface for inferring uni dynamic information including gene regulatory interaction and bursting kinetics from scRNA-seq data of two genes.
    """
    grn = pd.read_csv(grn_filename)
    counts = pd.read_csv(counts_filename)
    grn = np.matrix(grn)[:, 1::]
    counts = np.matrix(counts)[:, 2::]
    try:
        index1 = np.where(np.isin(grn[:, 0], genename1))[0]  
        index2 = np.where(np.isin(grn[:, 1], genename2))[0]   
        index = list(set(index1) & set(index2))[0]
        result = _global_uni_burst_link(grn_filename, counts_filename, index, verbose1, verbose2, verbose3, test)
        return(result)
    except IndexError:
        print('can not find genepair information')
    
def _global_uni_burst_link(grn_filename, counts_filename, index, verbose1, verbose2, verbose3, test):
    grn = pd.read_csv(grn_filename)
    counts = pd.read_csv(counts_filename)
    grn = np.matrix(grn)[:, 1::]
    counts = np.matrix(counts)[:, 2::]
    geneinfo, counts_data =  genepair_info(grn, counts, index)
    result_ = genepair_inference(counts_data, geneinfo, index, verbose1, verbose2, verbose3, test)
    result = np.hstack([geneinfo, result_.reshape([1, 20])])
    return(result)

def genepair_info(grn, counts, index):
    gene1 = grn[index, 0]
    gene2 = grn[index, 1]
    geneinfo = grn[index, :]
    index1 = list(counts[:, 0]).index(gene1)    
    index2 = list(counts[:, 0]).index(gene2)    
    counts_data = np.vstack([counts[index1, 1::].astype(float), counts[index2, 1::].astype(float)])  
    nan_indices1 = np.where(np.isnan(counts_data[0, :]))[1]
    nan_indices2 = np.where(np.isnan(counts_data[1, :]))[1]
    nan_indices = list(set(nan_indices1) | set(nan_indices2))
    counts_data = np.delete(counts_data, nan_indices, axis=1)
    counts_data = np.array(counts_data).astype(int)
    return (geneinfo, counts_data)

def genepair_inference(vals, geneinfo, figflag, verbose1, verbose2, verbose3, test):
    """
    Infer dynamic information including gene regulatory interaction and bursting kinetics from scRNA-seq data of two genes.

    Parameters
    ----------
    vals
        The data matrix of shape 2 × columns.
        Rows correspond to genes or spots and columns to cells.
    geneinfo
        The priori information from the base-GRN built from scATAC-seq data, containing names and regulatory types of two genes.
    figflag
        The index of the outputed images name.
        If choose not to output the images, the default value is 0.
    verbose1
        Whether to output images.
        If False, only the inferred results are outputed. If True, the inferred results with visualization images are outputed.
    verbose2
        Whether to save images.
        If False, the images are only shown in the runtime screen. If True, the images are saved as pdf format in a folder under the default path.
    verbose3
        Whether to exhibit all images.
        If 0, all images are exhibited. If 1, only visualization images are exhibited.
    test
        Whether to perform distributional test.
        If False, distribution test is not performed. If True, gibbs samples are generated and distribution test is performed.
    
    Returns
    -------
    np.array(['gene1', 'gene2', 'feedback1', 'feedback2', 'kon1', 'kon2', 'koff1', 'koff2', 'ksyn1', 'ksyn2', 'w', 'bf1', 'bf2', 'bs1', 'bs2', 'cv1', 'cv2', 'convergence', 'corr', 'NMI', 'cmi_given_x1', 'cmi_given_x2', 'sign_givenx1', 'sign_givenx2'])

   """
   
    truncated_data = pre_truncation(vals)
    bursting_results = genepair_burstinference(truncated_data)
    # bursting_results = [alpha1, alpha2, beta1, beta2, phi1, phi2, w, bf1, bf2, bs1, bs2, cv1, cv2, convergence_success]
    interactions_results = genepair_interactionsinference(truncated_data, geneinfo, 0.4, figflag, verbose1, verbose2, verbose3)
    # interactions_results = [corr, nmi, cmi_given_x1, cmi_given_x2, sign_givenx1, sign_givenx2]
    if (bursting_results[-1] == 1): Binary_PoBe_Interactions(bursting_results[0: 7], truncated_data, figflag, geneinfo[0, 0], geneinfo[0, 1], verbose1, verbose2, verbose3)
    if (test == True):
        if (bursting_results[-1] == 1): ks_distance, h0 = distribution_fit(truncated_data, interactions_results[0: 7])
        elif (bursting_results[-1] == 0): ks_distance, h0 = np.array([0, 0])
        return (np.hstack([bursting_results, interactions_results, np.array(ks_distance, h0)]))
    else: return (np.hstack([bursting_results, interactions_results]))

def pre_truncation(vals):
    boundary_value1 = round(np.mean(vals[0, :]) + 2 * np.std(vals[0, :]))
    boundary_value2 = round(np.mean(vals[1, :]) + 2 * np.std(vals[1, :]))
    indices1 = np.where(vals[0, :] > boundary_value1)[0]
    indices2 = np.where(vals[1, :] > boundary_value2)[0]
    indices = list(set(indices1) & set(indices2))
    truncated_vals = np.delete(vals, indices, axis=1)
    # Select the HDR of the data
    hdr_data = select_hdr(truncated_vals)
    return (hdr_data)

def select_hdr(data):
    kde = gaussian_kde(data)
    density = kde(data)
    # Identify data points with density above the threshold
    density_threshold = np.sort(density)[round(0.01*len(data[0, :]))]
    hdr_indices = np.where(density >= density_threshold)[0]
    hdr_data = data[:, hdr_indices]
    return (hdr_data)

def genepair_burstinference(data):
    """
    Infer bursting kinetics for each gene.
    """
    params0 = initialization(data)
    bnds = ((1e-4, 1e3),)*4 + ((0, 1e4),)*2 + ((-1e3, 1e3),)
    # Constraint conditions
    cons = ({'type': 'ineq', 'fun': lambda params: (params[0] + params[2]) * (params[1] + params[3]) / max(params[0] * params[3], params[1] * params[2]) - params[6]},
            {'type': 'ineq', 'fun': lambda params: (params[0] + params[2]) * (params[1] + params[3]) / max(params[0] * params[1], params[2] * params[3]) + params[6]})
    vals_ = np.copy(data) 
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category = RuntimeWarning)
        ll = minimize(log_likelihood, params0, args = (vals_), method = 'SLSQP', constraints = cons, bounds = bnds)
    estim = ll.x
    success = ll.success.astype(str)
    if (success == 'True'): convergence = 1
    else: convergence = 0
    alpha1, alpha2, beta1, beta2, phi1, phi2, w = estim[0: 7]
    bf1 = 1 / (1 / estim[0] + 1 / estim[2])
    bf2 = 1 / (1 / estim[1] + 1 / estim[3])
    bs1 = estim[4] / estim[2]
    bs2 = estim[5] / estim[3]
    cv1 = np.var(data[0, :]) / ((np.mean(data[0, :]))**2)
    cv2 = np.var(data[1, :]) / ((np.mean(data[1, :]))**2)
    infer_results = np.array([alpha1, alpha2, beta1, beta2, phi1, phi2, w, bf1, bf2, bs1, bs2, cv1, cv2, convergence])
    return (infer_results)

def uni_poisson(vals, m):
    if (max(m) < 1e6): return (poisson.pmf(vals, m))
    else: return (norm.pdf(vals, loc=m, scale=math.sqrt(m)))

def uni_pobe(vals, alpha_, beta_, phi_):
    vals = vals.reshape([len(vals), 1])
    x, w = j_roots(50, beta_ - 1, alpha_ - 1)
    gs_ = np.sum(w * uni_poisson(vals, phi_ * (1 + x) / 2), axis=1)
    prob = 1 / (beta_fun(alpha_, beta_)+1e-12) * 2 ** (-alpha_ - beta_ + 1) * gs_
    return (prob)

def uni_moment_inference(vals):
    m1 = float(np.mean(vals))
    m2 = float(sum(vals * (vals - 1)) / len(vals))
    m3 = float(sum(vals * (vals - 1) * (vals - 2)) / len(vals))
    r1 = m1
    r2 = m2 / m1
    r3 = m3 / m2
    alpha_est = (2*r1*(r3-r2)) / (r1*r2-2*r1*r3+r2*r3)
    beta_est = (2*(r3-r2)*(r1-r3)*(r2-r1)) / ((r1*r2-2*r1*r3+r2*r3) * (r1-2*r2+r3))
    phi_est = (2*r1*r3-r1*r2-r2*r3) / (r1-2*r2+r3)
    return (np.array([alpha_est, beta_est, phi_est]))

def uni_loglikelihood(params, vals):
    alpha_, beta_, phi_ = params[0: 3]
    prob = uni_pobe(vals, alpha_, beta_, phi_)
    uni_loglikelihood_ = -np.sum(np.log(prob + 1e-12))
    return (uni_loglikelihood_)
  
def uni_maximum_likelihood(vals):
    x0 = uni_moment_inference(vals)
    bnds = ((1e-3, 1e3), (1e-3, 1e3), (1, 1e4))
    vals_ = np.copy(vals)  
    try: ll = minimize(uni_loglikelihood, x0, args=(vals_), method='L-BFGS-B', bounds=bnds)
    except: return np.array([np.nan, np.nan, np.nan])
    return (ll.x)

def initialization(vals):
    mle1 = uni_maximum_likelihood(vals[0, :])
    mle2 = uni_maximum_likelihood(vals[1, :])
    alpha1, beta1, phi1 = mle1[0: 3]
    alpha2, beta2, phi2 = mle2[0: 3]
    cov = np.cov(vals[0], vals[1])[0, 1]
    w = float((((alpha1 + beta1)**2) * ((alpha2 + beta2)**2) * (alpha1 + beta1 + 1) * (alpha2 + beta2 + 1) * cov) / (phi1 * phi2 * alpha1 * alpha2 * beta1 * beta2))  
    return (np.array([alpha1, alpha2, beta1, beta2, phi1, phi2, w]))

def log_likelihood(params, data):
    alpha1, alpha2 = params[0: 2]
    beta1, beta2 = params[2: 4]
    phi1, phi2 = params[4: 6]
    w = params[6]
    alpha1_ = alpha1 + 1
    alpha2_ = alpha2 + 1
    mu1 = alpha1 / (alpha1 + beta1)
    mu2 = alpha2 / (alpha2 + beta2)
    prob1 = uni_pobe(data[0, :], alpha1, beta1, phi1)
    prob2 = uni_pobe(data[1, :], alpha2, beta2, phi2)
    prob1_ = uni_pobe(data[0, :], alpha1_, beta1, phi1)
    prob2_ = uni_pobe(data[1, :], alpha2_, beta2, phi2)
    a = w * mu1 * mu2
    prob = prob1 * prob2 + a * (prob1_ - prob1) * (prob2_ - prob2)
    indices = np.where(prob < 1e-12)[0]
    prob_ = prob[indices] = 1.0
    result = (-np.log(prob_)).sum()
    return (result)

def genepair_interactionsinference(data, geneinfo, filter_eps, figflag, verbose1, verbose2, verbose3):
    """
    Infer gene regulatory interactions and regulation visualization.

    Parameters
    ----------
    data
        The data matrix of shape 2 × columns.
        Rows correspond to genes or spots and columns to cells.
    geneinfo
        The priori information from the base-GRN built from scATAC-seq data, containing names and regulatory types of two genes.
    filter_eps
        DREMI filter noise threshold and default value is 0.4
    figflag
        The index of the outputed images name.
        If choose not to output the images, the default value is 0.
    verbose1
        Whether to output images.
        If False, only the inferred results are outputed. If True, the inferred results with visualization images are outputed.
    verbose2
        Whether to save images.
        If False, the images are only shown in the runtime screen. If True, the images are saved as pdf format in a folder under the default path.
    verbose3
        Whether to  exhibit all images.
        If 0, all images are exhibited. If 1, only visualization images are exhibited.
    
    Returns
    -------
    np.array([corr, nmi, cmi_given_x1, cmi_given_x2, sign_givenx1, sign_givenx2])
    
   """
   
    density = density_estimation(data)
    if (verbose2 == True): heatmap(density, figflag, 1, verbose2, geneinfo[0, 0], geneinfo[0, 1])
    # Calculate conditional probability
    px2_given_x1 = density / np.sum(density, axis=0)
    px1_given_x2 = density.T / np.sum(density.T, axis=0)
    rescale_density_given_x1, sign_given_x1 = rescale_density(px2_given_x1, float(geneinfo[0, 2]), figflag, 2, verbose1, verbose2, verbose3, True, geneinfo[0, 0], geneinfo[0, 1])
    rescale_density_given_x2, sign_given_x2 = rescale_density(px1_given_x2, float(geneinfo[0, 3]), figflag, 3, verbose1, verbose2, verbose3, True, geneinfo[0, 1], geneinfo[0, 0])
    # Regulaotry interactions
    sample_given_x1 = resample(rescale_density_given_x1, filter_eps)
    sample_given_x2 = resample(rescale_density_given_x2, filter_eps)
    cmi_given_x1 = conditional_mutualinfo(sample_given_x1, figflag, 4, verbose1, verbose2, verbose3)
    cmi_given_x2 = conditional_mutualinfo(sample_given_x2, figflag, 5, verbose1, verbose2, verbose2)
    corr = np.corrcoef(data[0, :], data[1, :])[0, 1]
    nmi = normalized_mutual_info_score(data[0, :], data[1, :])
    results = np.array([corr, nmi, cmi_given_x1, cmi_given_x2, sign_given_x1, sign_given_x2])
    return (results)

def density_estimation(data):
    kde = gaussian_kde(data, bw_method='scott')
    x_max = np.max(data[0, :])
    y_max = np.max(data[1, :])
    x_grid, y_grid = np.mgrid[0: x_max: 1, 0: y_max: 1]
    grid_points = np.vstack([x_grid.ravel(), y_grid.ravel()])
    density_ = kde(grid_points)
    density = density_ / np.sum(density_)
    density = density.reshape([x_max, y_max]).T
    return (density)

def heatmap(vals, figflag, figindex, verbose2, genename1, genename2):
    fig, ax = plt.subplots(dpi=900)
    ax.set_title('Joint Probability distrbution of data')
    im = ax.pcolormesh(vals, cmap=plt.cm.coolwarm, alpha=0.9)
    cb = fig.colorbar(im)
    plt.xlabel(genename1)
    plt.ylabel(genename2)
    if (verbose2 == True):
        plt.savefig(str(figflag) + '-' + str(figindex) + '.pdf')
        plt.close()
        plt.clf()
    return

def rescale_density(density, info, figflag, figindex, verbose1, verbose2, verbose3, regulation_type, genename1, genename2):
    density_max = np.max(density, axis=0)
    nn = density.shape[1]
    indices = np.where(density_max == 0)[0]
    if (len(indices) > 0):
        for index in indices:
            if (index == nn - 1): density[:, index] = density[:, index - 1]
            else: density[:, index] = (density[:, index - 1] + density[:, index + 1]) / 2
        density_max = np.max(density, axis=0)
        nn = density.shape[1]       
    rescaled_density = np.zeros(density.shape)
    for col in range(nn):
        rescaled_density[:, col] = density[:, col] / density_max[col]  
    if (verbose1 == True) & ((verbose3 == 0) | (verbose3 == 1)):
        heatmap(rescaled_density, figflag, figindex, verbose2, genename1, genename2)
    # Identification for regulation type
    if (regulation_type == True):
        col = rescaled_density.shape[1]
        samples = np.array(['x', 'y']).reshape([2, 1])
        for index in np.arange(col):
            indices = np.where(rescaled_density[:, index] >= 0.5)[0]
            values = np.repeat(index, len(indices))
            samples_ = np.vstack([values, indices])
            samples = np.hstack([samples, samples_])
        samples = samples[:, 1::].astype(float)
        coef, p_value = spearmanr(samples[0, :], samples[1, :])
        if (info == 0) & (p_value >= 0.99): sign_type = 0
        else: sign_type = -1 * (coef < 0) + 1 * (coef > 0)
        return (rescaled_density, sign_type)
    else: return (rescaled_density)

def rejected_samples(prob, num_samples):
    if (np.sum(prob) == 0):
        sample = []
    else:
        probabilities_ = prob / np.sum(prob)
        values = np.where(probabilities_ > 0)[0] 
        probabilities = probabilities_[values]
        samples = []
        while len(samples) < num_samples:
            sample = random.choices(values, probabilities)[0]
            samples.append(sample)
    return (samples)

def resample(density, filter_eps):
    filtered_density = np.where(density > filter_eps, density, 0)
    row, col = np.shape(filtered_density)
    num_samples = 20
    xx = []
    yy = []
    for index in range(col):
        y = rejected_samples(filtered_density[:, index], num_samples)
        if (len(y) == num_samples):
            x = np.repeat(index, num_samples).reshape((num_samples, 1))
            xx.append(x)
            yy.append(y)
    xx = np.array(xx).flatten().T
    yy = np.array(yy).flatten().T
    re_samples = np.vstack((xx, yy))
    return (re_samples)

def density_counts(vals):
    xmax = np.max(vals[0])
    ymax = np.max(vals[1])
    count = np.zeros([xmax, ymax])
    col = vals.shape[1]
    for i in range(xmax):
        for j in range(ymax):
            cnt = 0
            for k in range(col):
                if i <= vals[0, k] and vals[0, k] < i + 1 and j <= vals[1, k] and vals[1, k] < j + 1:
                    cnt = cnt + 1
            count[i, j] = cnt
    count = np.transpose(count)       
    pxy = count / np.sum(np.sum(count))
    return(pxy)

def conditional_mutualinfo(data, figflag, figindex, verbose1, verbose2, verbose3):
    density = density_counts(data)
    px = np.sum(density, axis=0)
    py = np.sum(density.T, axis=0)
    py_givenx = density / px
    rescaled_prob = rescale_density(py_givenx, 0, figflag, figindex, verbose1, verbose2, 3, False, 'X1', 'X2')
    # Calculate CMI
    cmi_ = np.zeros(density.shape)
    for col in np.arange(len(px)):
        for row in np.arange(len(py)):
            if (density[row, col] == 0 or px[col] == 0 or py[row] == 0):
                cmi_[row, col] = 0
            else:
                cmi_[row, col] = density[row, col] / px[col]  * (np.log(density[row, col]) - np.log(px[col]) - np.log(py[row]))
    cmi = np.sum(np.sum(cmi_))
    return(cmi)

def Binary_PoBe_Interactions(params, vals, figflag, genename1, genename2, verbose1, verbose2, verbose3): 
    alpha1, alpha2, beta1, beta2, phi1, phi2, w = params[0: 7]
    mu1 = alpha1 / (alpha1 + beta1)
    mu2 = alpha2 / (alpha2 + beta2)
    alpha1_ = alpha1 + 1
    alpha2_ = alpha2 + 1
    density = density_estimation(vals)
    col = vals[0, :].max()
    row = vals[1, :].max()
    xx, yy = np.mgrid[0:col:1, 0:row:1]
    vals = np.vstack([xx.ravel(), yy.ravel()])
    n = np.shape(vals)[1]
    prob1 = uni_pobe(vals[0, :], alpha1, beta1, phi1)
    prob2 = uni_pobe(vals[1, :], alpha2, beta2, phi2)
    prob1_ = uni_pobe(vals[0, :], alpha1_, beta1, phi1)
    prob2_ = uni_pobe(vals[1, :], alpha2_, beta2, phi2)
    pxy = []
    py_givenx = []
    px_giveny = []
    for index in range(n):
        pxy_ = prob1[index] * prob2[index] + w * mu1 * mu2 * (prob1_[index] - prob1[index]) * (prob2_[index] - prob2[index])
        py_givenx_ = prob2[index] + w * mu1 * mu2 * (prob2_[index] - prob2[index]) * (prob1_[index] / prob1[index] - 1)
        px_giveny_ = prob1[index] + w * mu1 * mu2 * (prob1_[index] - prob1[index]) * (prob2_[index] / prob2[index] - 1)
        pxy.append(pxy_)
        px_giveny.append(px_giveny_)
        py_givenx.append(py_givenx_)
    pxy = np.array(pxy).reshape([col, row]).T
    py_givenx = np.array(py_givenx).reshape([col, row]).T
    px_giveny = np.array(px_giveny).reshape([col, row])
    rescale_py_givenx = rescale_density(py_givenx, 0, figflag, 0, False, False, 3, False, 'genename1', 'genename2')
    rescale_px_giveny = rescale_density(px_giveny, 0, figflag, 0, False, False, 3, False, 'genename2', 'genename1')
    px = np.sum(pxy, axis=0)
    py = np.sum(pxy, axis=1)
    if (verbose1 == True) & ((verbose3 == 0) | (verbose3 == 2)):
        if (verbose2 == False):
            fig, axes = plt.subplots(1, 3, figsize=(9, 3), dpi=300) 
            axes[0].pcolormesh(density, cmap=plt.cm.coolwarm, alpha=0.9)   
            axes[0].set_title("Joint Probability distrbution", fontsize=7)
            axes[0].set_xlabel(genename1)
            axes[0].set_ylabel(genename2)
            axes[1].pcolormesh(rescale_py_givenx, cmap=plt.cm.coolwarm, alpha=0.9)   
            axes[1].set_title("Rescaled Conditional Probability Given $X_1$", fontsize=7)
            axes[1].set_xlabel(genename1)
            axes[1].set_ylabel(genename2)
            axes[2].pcolormesh(rescale_px_giveny, cmap=plt.cm.coolwarm, alpha=0.9)
            axes[2].set_title("Rescaled Conditional Probability Given $X_2$", fontsize=7)
            axes[2].set_xlabel(genename2)
            axes[2].set_ylabel(genename1)
            plt.tight_layout()
            plt.show()
        else:
            fig1, ax1 = plt.subplots(dpi=900)
            ax1.set_title('Theory Inference Rescaled Conditional Probability Given $X_1$')
            im1 = ax1.pcolormesh(rescale_py_givenx, cmap=plt.cm.coolwarm, alpha=0.9)
            fig1.colorbar(im1, ax=ax1)
            plt.xlabel(genename1)
            plt.ylabel(genename2)
            plt.savefig('Rescaled conditional probability1.pdf')  
            plt.close()
            plt.clf()
            fig2, ax2 = plt.subplots(dpi=900)
            ax2.set_title('Theory Inference Rescaled Conditional Probability Given $X_2$')
            im2 = ax2.pcolormesh(rescale_px_giveny, cmap=plt.cm.coolwarm, alpha=0.9)
            fig2.colorbar(im2, ax=ax2)
            plt.xlabel(genename2)
            plt.ylabel(genename1)
            plt.savefig('Rescaled conditional probability2.pdf')  
            plt.close()
            plt.clf() 
    return
    
def distribution_fit(counts_data, stat_params):
    m = 500
    n = 500 + m + 1
    stat_data = _synthetic_data.gibbs_sample(np.array(stat_params), m, n)
    #  KW test
    data1 = np.transpose(counts_data[:, -500:])
    data2 = np.transpose(stat_data)
    S, KS_critical_val, H, p_value = ks_2samp(data1, data2, alpha=0.05)
    return(S, H)

def ks_2samp(x_val, y_val, alpha, asymptotic = False):
    """
    Performs a multivariate two-sample extension of the Kolmogorov-Smirnov test.

    Parameters
    ----------
    x_val
        A numpy array of shape (num_samples_x, dim) representing the first sample.
    y_val
        A numpy array of shape (num_samples_y, dim) representing the second sample.
    alpha
        The significance level.
    asymptotic
        Whether to use the asymptotic approximation or not.
    
    Returns
    -------
    A boolean indicating whether the null hypothesis is rejected.
    
   """
   
    num_samples_x, dim = x_val.shape
    num_samples_y, num_feats_y = y_val.shape
    if dim != num_feats_y:
        raise ValueError("The two samples do not have the same number of features.")
    z = np.zeros((num_samples_x, dim, dim))
    for h in range(dim):
        ind = np.argsort(x_val[:, h])[::-1]
        temp = np.take(x_val, ind, axis=0)
        z[:, :, h] = temp
        for i in range(dim):
            for j in range(num_samples_x - 1, -1, -1):
                if j == num_samples_x - 1:
                    runmax = temp[num_samples_x - 1, i]
                else:
                    runmax = max(runmax, temp[j, i])
                z[j, i, h] = runmax
    diff = np.zeros((num_samples_x, dim))
    for h in range(dim):
        for i in range(num_samples_x):
            val = np.abs(mecdf(x_val, z[i, :, h]) - mecdf(y_val, z[i, :, h])) * (
                round(num_samples_x * mecdf(x_val, z[i, :, h])) == num_samples_x - i
            )
            diff[i, h] = val
            if h == 0:
                diff[i, h] = max(
                    diff[i, h],
                    np.abs(mecdf(x_val, x_val[i, :]) - mecdf(y_val, x_val[i, :])),
                )
    KS = np.max(diff)
    if asymptotic:
        KS_critical_val = np.sqrt(-np.log(alpha / (4 * dim)) * (0.5 / num_samples_x)) + np.sqrt(
            (-1) * np.log(alpha / (4 * dim)) * (0.5 / num_samples_y)
        )
    else:
        KS_critical_val = np.sqrt(
            -np.log(alpha / (2 * (num_samples_x + 1) * dim)) * (0.5 / num_samples_x)
        ) + np.sqrt((-1) * np.log(alpha / (2 * (num_samples_y + 1) * dim)) * (0.5 / num_samples_y))
        
    if (KS > KS_critical_val):
        H = 'True'
        p_value = 1
    else:
        H = 'False'
        phi = -stats.norm.isf(1-alpha)
        sigma = KS_critical_val / phi
        p_value = 1 - stats.norm.sf(x = ((KS- 2*KS_critical_val) / sigma), loc = 0, scale = 1)
    return (KS, KS_critical_val, H, p_value)

def mecdf(x_val, t):
    """
    Computes the multivariate empirical cdf of x_val at t.

    Parameters
    ----------
    x_val
        A numpy array of shape (num_samples_x, dim) representing the sample.
    t
        A numpy array of shape (num_samples_t, dim) representing the point at which to evaluate the cdf.
    
    Returns
    -------
    The multivariate empirical cdf of x_val at t.  
    """

    lower = (x_val <= t) * 1.0
    return (np.mean(np.prod(lower, axis=1)))

def Binary_PoBe(params, vals): 
    alpha1, alpha2, beta1, beta2, phi1, phi2, w = params[0: 7]
    mu1 = alpha1 / (alpha1 + beta1)
    mu2 = alpha2 / (alpha2 + beta2)
    alpha1_ = alpha1 + 1
    alpha2_ = alpha2 + 1
    col = vals[0, :].max()
    row = vals[1, :].max()
    xx, yy = np.mgrid[0:col:1, 0:row:1]
    vals = np.vstack([xx.ravel(), yy.ravel()])
    n = np.shape(vals)[1]
    prob1 = uni_pobe(vals[0, :], alpha1, beta1, phi1)
    prob2 = uni_pobe(vals[1, :], alpha2, beta2, phi2)
    prob1_ = uni_pobe(vals[0, :], alpha1_, beta1, phi1)
    prob2_ = uni_pobe(vals[1, :], alpha2_, beta2, phi2)
    pxy = []
    py_givenx = []
    px_giveny = []
    for index in range(n):
        pxy_ = prob1[index] * prob2[index] + w * mu1 * mu2 * (prob1_[index] - prob1[index]) * (prob2_[index] - prob2[index])
        py_givenx_ = prob2[index] + w * mu1 * mu2 * (prob2_[index] - prob2[index]) * (prob1_[index] / prob1[index] - 1)
        px_giveny_ = prob1[index] + w * mu1 * mu2 * (prob1_[index] - prob1[index]) * (prob2_[index] / prob2[index] - 1)
        pxy.append(pxy_)
        px_giveny.append(px_giveny_)
        py_givenx.append(py_givenx_)
    pxy = np.array(pxy).reshape([col, row]).T
    py_givenx = np.array(py_givenx).reshape([col, row]).T
    px_giveny = np.array(px_giveny).reshape([col, row])
    px = np.sum(pxy, axis=0)
    py = np.sum(pxy, axis=1)
    return (pxy, px, py)


from setuptools import setup, find_packages
from pathlib import Path

long_description = Path("README.md").read_text(encoding="utf-8")

setup(
    name="burstlink",                  
    version="1.0.1",   
    description="A user-friendly package for analyzing gene interactions and transcriptional bursting.",                
    packages=find_packages(),            
    python_requires=">=3.8",   
    install_requires=[
        "statsmodels>=0.12.0",
        "numpy>=1.21.0",
        "pandas>=1.3.0",
    ],
    long_description=long_description,
    long_description_content_type="text/markdown",  
)